# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
cimport numpy as np
cimport openmp as omp
from cython.parallel import parallel, prange
from libc.math cimport exp, log, sqrtf
from libc.stdint cimport uint8_t, uint32_t
from libc.stdlib cimport calloc, free

ctypedef uint8_t u8
ctypedef uint32_t u32
ctypedef float f32
ctypedef double f64

cdef f64 PRO_MIN = 1e-5
cdef f64 PRO_MAX = 1.0 - (1e-5)
cdef f64 ACC_MIN = 1.0
cdef f64 ACC_MAX = 128.0
cdef f32 FLT_MIN = 1e-5
cdef f32 FLT_MAX = 1.0 - (1e-5)
cdef inline f64 _fmax(f64 a, f64 b) noexcept nogil: return a if a > b else b
cdef inline f64 _fmin(f64 a, f64 b) noexcept nogil: return a if a < b else b
cdef inline f32 _fmaxf(f32 a, f32 b) noexcept nogil: return a if a > b else b
cdef inline f32 _fminf(f32 a, f32 b) noexcept nogil: return a if a < b else b
cdef inline f64 _clamp1(f64 a) noexcept nogil: return _fmax(PRO_MIN, _fmin(a, PRO_MAX))
cdef inline f64 _clamp2(f64 a) noexcept nogil: return _fmax(ACC_MIN, _fmin(a, ACC_MAX))
cdef inline f32 _clamp3(f32 a) noexcept nogil: return _fmaxf(FLT_MIN, _fminf(a, FLT_MAX))


##### hapla - ancestry estimation #####
### Inline functions
# Estimate inverse individual allele frequency
cdef inline f64 _computeH(
		const f64* p, const f64* q, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
		f64 h = 0.0
	for k in range(K):
		h += p[k]*q[k]
	return 1.0/h

# Estimate individual allele frequency
cdef inline f64 _computeL(
		const f64* p, const f64* q, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
		f64 h = 0.0
	for k in range(K):
		h += p[k]*q[k]
	return log(h)

# Inner loop updates for temp P and Q
cdef inline void _innerJ(
		const f64* p, const f64* q, f64* p_thr, f64* q_thr, const f64 h, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
	for k in range(K):
		p_thr[k] += q[k]*h
		q_thr[k] += p[k]*h

# Inner loop updates for temp Q
cdef inline void _innerQ(
		const f64* p, f64* q_thr, const f64 h, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
	for k in range(K):
		q_thr[k] += p[k]*h

# Outer loop update for P
cdef inline void _outerP(
		f64* p, f64* p_thr, const f64 S, const Py_ssize_t B, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t c, k
		f64 a, b
		f64* p_sum = <f64*>calloc(K, sizeof(f64))
		f64* p_c
		f64* p_t
	for c in range(B):
		p_c = &p[c*K]
		p_t = &p_thr[c*K]
		for k in range(K):
			a = p_c[k]*p_t[k]*S
			b = _clamp1(a)
			p_sum[k] += b
			p_c[k] = b
			p_t[k] = 0.0
	for c in range(B):
		p_c = &p[c*K]
		for k in range(K):
			p_c[k] /= p_sum[k]
	free(p_sum)

# Outer loop accelerated update for P
cdef inline void _outerAccelP(
		f64* p, f64* p_new, f64* p_thr, const f64 S, const Py_ssize_t B, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t c, k
		f64 a, b
		f64* p_sum = <f64*>calloc(K, sizeof(f64))
		f64* p_c
		f64* p_n
		f64* p_t
	for c in range(B):
		p_c = &p[c*K]
		p_n = &p_new[c*K]
		p_t = &p_thr[c*K]
		for k in range(K):
			a = p_c[k]*p_t[k]*S
			b = _clamp1(a)
			p_sum[k] += b
			p_n[k] = b
			p_t[k] = 0.0
	for c in range(B):
		p_n = &p_new[c*K]
		for k in range(K):
			p_n[k] /= p_sum[k]
	free(p_sum)

# Outer loop update for Q
cdef inline void _outerQ(
		f64* q, f64* q_tmp, const f64 S, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
		f64 sumQ = 0.0
		f64 a, b
	for k in range(K):
		a = q[k]*q_tmp[k]*S
		b = _clamp1(a)
		sumQ += b
		q[k] = b
		q_tmp[k] = 0.0
	for k in range(K):
		q[k] /= sumQ

# Outer loop accelerated update for Q
cdef inline void _outerAccelQ(
		const f64* q, f64* q_new, f64* q_tmp, const f64 S, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
		f64 sumQ = 0.0
		f64 a, b
	for k in range(K):
		a = q[k]*q_tmp[k]*S
		b = _clamp1(a)
		sumQ += b
		q_new[k] = b
		q_tmp[k] = 0.0
	for k in range(K):
		q_new[k] /= sumQ

# Estimate QN factor
cdef inline f64 _qnC(
		const f64* v0, const f64* v1, const f64* v2, const Py_ssize_t I
	) noexcept nogil:
	cdef:
		size_t i
		f64 sum1 = 0.0
		f64 sum2 = 0.0
		f64 f, u, v
	for i in prange(I, schedule='guided'):
		u = v1[i] - v0[i]
		v = v2[i] - v1[i] - u
		sum1 += u*u
		sum2 += u*v
	f = -(sum1/sum2)
	return _clamp2(f)

# Estimate batch QN factor for P
cdef inline f64 _qnBatch(
		f64* P0, f64* P1, f64* P2, const u32* k_vec, const u32* c_vec, const u32* s_bat, const Py_ssize_t W, 
		const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t c, l, r, w
		f64 sum1 = 0.0
		f64 sum2 = 0.0
		f64 f, u, v
		f64* p0
		f64* p1
		f64* p2
	for w in prange(W, schedule='guided'):
		r = s_bat[w]
		l = c_vec[r]
		p0 = &P0[l]
		p1 = &P1[l]
		p2 = &P2[l]
		for c in range(k_vec[r]*K):
			u = p1[c] - p0[c]
			v = p2[c] - p1[c] - u
			sum1 += u*u
			sum2 += u*v
	f = -(sum1/sum2)
	return _clamp2(f)

# Estimate QN jump in P
cdef inline void _computeP(
		f64* P0, f64* P1, f64* P2, const f64 c1, const f64 c2, const Py_ssize_t B, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t c, k
		f64 a, b
		f64* p_sum = <f64*>calloc(K, sizeof(f64))
		f64* p0
		f64* p1
		f64* p2
	for c in range(B):
		p0 = &P0[c*K]
		p1 = &P1[c*K]
		p2 = &P2[c*K]
		for k in range(K):
			a = c2*p1[k] + c1*p2[k]
			b = _clamp1(a)
			p_sum[k] += b
			p0[k] = b
	for c in range(B):
		p0 = &P0[c*K]
		for k in range(K):
			p0[k] /= p_sum[k]
	free(p_sum)

# Estimate QN jump in Q
cdef inline void _computeQ(
		f64* q0, const f64* q1, const f64* q2, const f64 c1, const f64 c2, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
		f64 sumQ = 0.0
		f64 a, b
	for k in range(K):
		a = c2*q1[k] + c1*q2[k]
		b = _clamp1(a)
		sumQ += b
		q0[k] = b
	for k in range(K):
		q0[k] /= sumQ

# Project P to domain
cdef inline void _projectP(
		f32* p, const Py_ssize_t B, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t c, k
		f32 a, b
		f32* p_sum = <f32*>calloc(K, sizeof(f32))
		f32* p_c
	for c in range(B):
		p_c = &p[c*K]
		for k in range(K):
			a = p_c[k]
			b = _clamp3(a)
			p_sum[k] += b
			p_c[k] = b
	for c in range(B):
		p_c = &p[c*K]
		for k in range(K):
			p_c[k] /= p_sum[k]
	free(p_sum)

# Project Q to domain
cdef inline void _projectQ(
		f32* q, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
		f32 sumQ = 0.0
		f32 a, b
	for k in range(K):
		a = q[k]
		b = _clamp3(a)
		sumQ += b
		q[k] = b
	for k in range(K):
		q[k] /= sumQ

# Compute the squared difference
cdef inline f32 _computeR(
		const f32* a, const f32* b, const Py_ssize_t I
	) noexcept nogil:
	cdef:
		f32 r = 0.0
		f32 c
		size_t i
	for i in range(I):
		c = a[i] - b[i]
		r += c*c
	return r


### Update functions
# Update P and Q temp arrays
cpdef void updateP(
		u8[:,::1] Z, f64[::1] P, const f64[:,::1] Q, f64[:,::1] Q_tmp, const u32[::1] k_vec, const u32[::1] c_vec, 
		const u32 L
	) noexcept nogil:
	cdef:
		Py_ssize_t W = Z.shape[0]
		Py_ssize_t N = Q.shape[0]
		Py_ssize_t K = Q.shape[1]
		Py_ssize_t B
		size_t i, l, s, w, x, y
		f64 S = 1.0/<f64>(2*N)
		f64 h
		f64* p
		f64* q
		f64* p_thr
		f64* q_thr
		omp.omp_lock_t mutex
	omp.omp_init_lock(&mutex)
	with nogil, parallel():
		p_thr = <f64*>calloc(L*K, sizeof(f64))
		q_thr = <f64*>calloc(N*K, sizeof(f64))
		for w in prange(W, schedule='guided'):
			l = c_vec[w]
			B = k_vec[w]
			for i in range(2*N):
				s = Z[w,i]*K
				p = &P[l + s]
				q = &Q[i >> 1,0]
				h = _computeH(p, q, K)
				_innerJ(p, q, &p_thr[s], &q_thr[(i >> 1)*K], h, K)
			_outerP(&P[l], &p_thr[0], S, B, K)

		# omp critical
		omp.omp_set_lock(&mutex)
		for x in range(N):
			for y in range(K):
				Q_tmp[x,y] += q_thr[x*K + y]
		omp.omp_unset_lock(&mutex)
		free(p_thr)
		free(q_thr)
	omp.omp_destroy_lock(&mutex)

# Accelerated update P and Q temp arrays
cpdef void accelP(
		u8[:,::1] Z, f64[::1] P, f64[::1] P_new, const f64[:,::1] Q, f64[:,::1] Q_tmp, const u32[::1] k_vec, 
		const u32[::1] c_vec, const u32 L
	) noexcept nogil:
	cdef:
		Py_ssize_t W = Z.shape[0]
		Py_ssize_t N = Q.shape[0]
		Py_ssize_t K = Q.shape[1]
		Py_ssize_t B
		size_t i, l, s, w, x, y
		f64 S = 1.0/<f64>(2*N)
		f64 h
		f64* p
		f64* q
		f64* p_thr
		f64* q_thr
		omp.omp_lock_t mutex
	omp.omp_init_lock(&mutex)
	with nogil, parallel():
		p_thr = <f64*>calloc(L*K, sizeof(f64))
		q_thr = <f64*>calloc(N*K, sizeof(f64))
		for w in prange(W, schedule='guided'):
			l = c_vec[w]
			B = k_vec[w]
			for i in range(2*N):
				s = Z[w,i]*K
				p = &P[l + s]
				q = &Q[i >> 1,0]
				h = _computeH(p, q, K)
				_innerJ(p, q, &p_thr[s], &q_thr[(i >> 1)*K], h, K)
			_outerAccelP(&P[l], &P_new[l], &p_thr[0], S, B, K)

		# omp critical
		omp.omp_set_lock(&mutex)
		for x in range(N):
			for y in range(K):
				Q_tmp[x,y] += q_thr[x*K + y]
		omp.omp_unset_lock(&mutex)
		free(p_thr)
		free(q_thr)
	omp.omp_destroy_lock(&mutex)

# Batch accelerated update P and Q temp arrays
cpdef void accelBatchP(
		u8[:,::1] Z, f64[::1] P, f64[::1] P_new, const f64[:,::1] Q, f64[:,::1] Q_tmp, const u32[::1] k_vec, 
		const u32[::1] c_vec, const u32[::1] s_bat, const u32 L
	) noexcept nogil:
	cdef:
		Py_ssize_t W = s_bat.shape[0]
		Py_ssize_t N = Q.shape[0]
		Py_ssize_t K = Q.shape[1]
		Py_ssize_t B
		size_t i, l, r, s, w, x, y
		f64 S = 1.0/<f64>(2*N)
		f64 h
		f64* p
		f64* q
		f64* p_thr
		f64* q_thr
		omp.omp_lock_t mutex
	omp.omp_init_lock(&mutex)
	with nogil, parallel():
		p_thr = <f64*>calloc(L*K, sizeof(f64))
		q_thr = <f64*>calloc(N*K, sizeof(f64))
		for w in prange(W, schedule='guided'):
			r = s_bat[w]
			l = c_vec[r]
			B = k_vec[r]
			for i in range(2*N):
				s = Z[r,i]*K
				p = &P[l + s]
				q = &Q[i >> 1,0]
				h = _computeH(p, q, K)
				_innerJ(p, q, &p_thr[s], &q_thr[(i >> 1)*K], h, K)
			_outerAccelP(&P[l], &P_new[l], &p_thr[0], S, B, K)

		# omp critical
		omp.omp_set_lock(&mutex)
		for x in range(N):
			for y in range(K):
				Q_tmp[x,y] += q_thr[x*K + y]
		omp.omp_unset_lock(&mutex)
		free(p_thr)
		free(q_thr)
	omp.omp_destroy_lock(&mutex)

# Accelerated jump for P (QN)
cpdef void jumpP(
		f64[::1] P0, f64[::1] P1, f64[::1] P2, const u32[::1] k_vec, const u32[::1] c_vec, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		Py_ssize_t M = P0.shape[0]
		Py_ssize_t W = k_vec.shape[0]
		Py_ssize_t B
		size_t l, w
		f64 c1, c2
	c1 = _qnC(&P0[0], &P1[0], &P2[0], M)
	c2 = 1.0 - c1
	with nogil, parallel():
		for w in prange(W, schedule='guided'):
			l = c_vec[w]
			B = k_vec[w]
			_computeP(&P0[l], &P1[l], &P2[l], c1, c2, B, K)

# Batch accelerated jump for P (QN)
cpdef void jumpBatchP(
		f64[::1] P0, const f64[::1] P1, const f64[::1] P2, const u32[::1] k_vec, const u32[::1] c_vec, 
		const u32[::1] s_bat, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		Py_ssize_t W = s_bat.shape[0]
		Py_ssize_t B
		size_t l, r, w
		f64 c1, c2
	c1 = _qnBatch(&P0[0], &P1[0], &P2[0], &k_vec[0], &c_vec[0], &s_bat[0], W, K)
	c2 = 1.0 - c1
	with nogil, parallel():
		for w in prange(W, schedule='guided'):
			r = s_bat[w]
			l = c_vec[r]
			B = k_vec[r]
			_computeP(&P0[l], &P1[l], &P2[l], c1, c2, B, K)

# Update Q
cpdef void updateQ(
		f64[:,::1] Q, f64[:,::1] Q_tmp, const Py_ssize_t W
	) noexcept nogil:
	cdef:
		Py_ssize_t N = Q.shape[0]
		Py_ssize_t K = Q.shape[1]
		size_t i, k
		f64 S = 1.0/<f64>(2*W)
	for i in prange(N, schedule='guided'):
		_outerQ(&Q[i,0], &Q_tmp[i,0], S, K)

# Accelerated update Q
cpdef void accelQ(
		const f64[:,::1] Q, f64[:,::1] Q_new, f64[:,::1] Q_tmp, const Py_ssize_t W
	) noexcept nogil:
	cdef:
		Py_ssize_t N = Q.shape[0]
		Py_ssize_t K = Q.shape[1]
		size_t i, k
		f64 S = 1.0/<f64>(2*W)
	for i in prange(N, schedule='guided'):
		_outerAccelQ(&Q[i,0], &Q_new[i,0], &Q_tmp[i,0], S, K)

# Accelerated jump for Q (QN)
cpdef void jumpQ(
		f64[:,::1] Q0, const f64[:,::1] Q1, const f64[:,::1] Q2
	) noexcept nogil:
	cdef:
		Py_ssize_t N = Q0.shape[0]
		Py_ssize_t K = Q0.shape[1]
		size_t i, k
		f64 a, c1, c2
	c1 = _qnC(&Q0[0,0], &Q1[0,0], &Q2[0,0], N*K)
	c2 = 1.0 - c1
	for i in prange(N, schedule='guided'):
		_computeQ(&Q0[i,0], &Q1[i,0], &Q2[i,0], c1, c2, K)


### Other functions
# Create P matrix from array
cpdef void createP(
		f64[::1] P, const u32[::1] k_vec, const u32[::1] c_vec, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		Py_ssize_t W = k_vec.shape[0]
		Py_ssize_t B
		size_t c, k, l, w
		f64* p_sum
	with nogil, parallel():
		p_sum = <f64*>calloc(K, sizeof(f64))
		for w in prange(W, schedule='guided'):
			l = c_vec[w]
			B = k_vec[w]
			for c in range(B):
				p = &P[l + c*K]
				for k in range(K):
					p_sum[k] += p[k]
			for c in range(B):
				p = &P[l + c*K]
				for k in range(K):
					p[k] /= p_sum[k]
			for k in range(K):
				p_sum[k] = 0.0
		free(p_sum)

# Log-likelihood
cpdef f64 loglike(
		u8[:,::1] Z, f64[::1] P, const f64[:,::1] Q, const u32[::1] c_vec
	) noexcept nogil:
	cdef:
		Py_ssize_t W = Z.shape[0]
		Py_ssize_t N = Z.shape[1]
		Py_ssize_t K = Q.shape[1]
		Py_ssize_t M = P.shape[0]
		size_t i, l, w
		f64 r = 0.0
		f64 h
		f64* p
	for w in prange(W, schedule='guided'):
		l = c_vec[w]
		p = &P[l]
		for i in range(N):
			r += _computeL(&p[Z[w,i]*K], &Q[i >> 1,0], K)
	return r*((<f64>K)/((<f64>M)*(<f64>N)))

# Projection function for P (f32)
cpdef void projectP(
		f32[:,::1] P, const u32[::1] k_vec, const u32[::1] c_vec
	) noexcept nogil:
	cdef:
		Py_ssize_t W = k_vec.shape[0]
		Py_ssize_t K = P.shape[1]
		Py_ssize_t B
		size_t l, w
	with nogil, parallel():
		for w in prange(W, schedule='guided'):
			l = c_vec[w]
			B = k_vec[w]
			_projectP(&P[l,0], B, K)

# Projection function for Q (f32)
cpdef void projectQ(
		f32[:,::1] Q
	) noexcept nogil:
	cdef:
		Py_ssize_t N = Q.shape[0]
		Py_ssize_t K = Q.shape[1]
		size_t i
	for i in prange(N, schedule='guided'):
		_projectQ(&Q[i,0], K)

# Root-mean square error between two Q matrices
cpdef f32 rmseQ(
		f32[:,::1] A, f32[:,::1] B
	) noexcept nogil:
	cdef:
		Py_ssize_t N = A.shape[0]
		Py_ssize_t K = A.shape[1]
		f32 r
	r = _computeR(&A[0,0], &B[0,0], N*K)
	return sqrtf(r/(<f32>(N)*<f32>(K)))


### Supervised mode
# Initialize P based on Q and haplotype cluster assignments
cpdef void superP(
		u8[:,::1] Z, f64[:,::1] P, const u32[::1] k_vec, const u32[::1] c_vec, u8[::1] z
	) noexcept nogil:
	cdef:
		Py_ssize_t W = Z.shape[0]
		Py_ssize_t N = Z.shape[1]
		Py_ssize_t K = P.shape[1]
		Py_ssize_t B
		size_t c, k, i, l, w
		u8 y
		f64 a, b
		f64* p
		f64* c_sum
		f64* p_sum
	with nogil, parallel():
		c_sum = <f64*>calloc(K, sizeof(f64))
		p_sum = <f64*>calloc(K, sizeof(f64))
		for w in prange(W, schedule='guided'):
			l = c_vec[w]
			B = k_vec[w]
			for i in range(N):
				y = z[i >> 1]
				if y > 0:
					P[l + Z[w,i], y - 1] += 1.0
					c_sum[y - 1] += 1.0
			for c in range(B):
				p = &P[l + c,0]
				for k in range(K):
					a = p[k]
					if c_sum[k] > 0.0:
						a = a/c_sum[k]
					b = _clamp1(a)
					p_sum[k] += b
					p[k] = b
			for c in range(B):
				p = &P[l + c,0]
				for k in range(K):
					p[k] /= p_sum[k]
			for k in range(K):
				c_sum[k] = 0.0
				p_sum[k] = 0.0
		free(c_sum)
		free(p_sum)

# Update Q in supervised mode
cpdef void superQ(
		f64[:,::1] Q, const u8[::1] y
	) noexcept nogil:
	cdef:
		Py_ssize_t K = Q.shape[1]
		Py_ssize_t N = Q.shape[0]
		size_t i, k
		f64 sumQ
	for i in prange(N, schedule='guided'):
		if y[i] > 0:
			sumQ = 0.0
			for k in range(K):
				if k == (y[i] - 1):
					Q[i,k] = PRO_MAX
				else:
					Q[i,k] = PRO_MIN
				sumQ = sumQ + Q[i,k]
			for k in range(K):
				Q[i,k] /= sumQ


### Projection mode
# Check haplotype cluster frequencies input
cpdef void checkP(
		f64[::1] P, f64[:,::1] p_sum, const u32[::1] k_vec, const u32[::1] c_vec, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		Py_ssize_t W = k_vec.shape[0]
		Py_ssize_t B
		f64* p
		size_t c, k, l, w
	for w in prange(W, schedule='guided'):
		l = c_vec[w]
		B = k_vec[w]
		for c in range(B):
			p = &P[l + c*K]
			for k in range(K):
				p_sum[w,k] += p[k]

# Update Q temp arrays in projection mode
cpdef void stepQ(
		u8[:,::1] Z, f64[::1] P, const f64[:,::1] Q, f64[:,::1] Q_tmp, const u32[::1] k_vec, const u32[::1] c_vec
	) noexcept nogil:
	cdef:
		Py_ssize_t W = Z.shape[0]
		Py_ssize_t N = Q.shape[0]
		Py_ssize_t K = Q.shape[1]
		size_t i, l, s, w, x, y
		f64 h
		f64* p
		f64* q_thr
		omp.omp_lock_t mutex
	omp.omp_init_lock(&mutex)
	with nogil, parallel():
		q_thr = <f64*>calloc(N*K, sizeof(f64))
		for w in prange(W, schedule='guided'):
			l = c_vec[w]
			for i in range(2*N):
				s = l + Z[w,i]*K
				p = &P[s]
				h = _computeH(p, &Q[i >> 1,0], K)
				_innerQ(p, &q_thr[(i >> 1)*K], h, K)

		# omp critical
		omp.omp_set_lock(&mutex)
		for x in range(N):
			for y in range(K):
				Q_tmp[x,y] += q_thr[x*K + y]
		omp.omp_unset_lock(&mutex)
		free(q_thr)
	omp.omp_destroy_lock(&mutex)

# Batch accelerate update Q temp arrays in projection mode
cpdef void stepBatchQ(
		u8[:,::1] Z, f64[::1] P, const f64[:,::1] Q, f64[:,::1] Q_tmp, const u32[::1] k_vec, const u32[::1] c_vec, 
		const u32[::1] s_bat
	) noexcept nogil:
	cdef:
		Py_ssize_t W = s_bat.shape[0]
		Py_ssize_t N = Q.shape[0]
		Py_ssize_t K = Q.shape[1]
		size_t i, l, r, s, w, x, y
		f64 h
		f64* p
		f64* q_thr
		omp.omp_lock_t mutex
	omp.omp_init_lock(&mutex)
	with nogil, parallel():
		q_thr = <f64*>calloc(N*K, sizeof(f64))
		for w in prange(W, schedule='guided'):
			r = s_bat[w]
			l = c_vec[r]
			for i in range(2*N):
				s = l + Z[r,i]*K
				p = &P[s]
				h = _computeH(p, &Q[i >> 1,0], K)
				_innerQ(p, &q_thr[(i >> 1)*K], h, K)

		# omp critical
		omp.omp_set_lock(&mutex)
		for x in range(N):
			for y in range(K):
				Q_tmp[x,y] += q_thr[x*K + y]
		omp.omp_unset_lock(&mutex)
		free(q_thr)
	omp.omp_destroy_lock(&mutex)



# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
cimport numpy as np
from cython.parallel import prange
from libc.math cimport exp, log
from libc.stdint cimport uint8_t, uint32_t

ctypedef uint8_t u8
ctypedef uint32_t u32
ctypedef float f32
ctypedef double f64


##### hapla - local ancestry inference #####
# Safe log-sum-exp for array
cdef inline f64 _logsumexp(
		const f64* vec, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
		f64 max_v = vec[0]
		f64 sum_v = 0.0
	for k in range(1, K):
		if vec[k] > max_v:
			max_v = vec[k]
	for k in range(K):
		sum_v += exp(vec[k] - max_v)
	return log(sum_v) + max_v

# Inner normalization function for likelihoods
cdef inline void _normLikes(
		f32* l, const u32 B
	) noexcept nogil:
	cdef:
		size_t z, c1, c2
		f32 sumC, tmpC
	for c1 in range(B):
		z = c1*B
		sumC = 0.0
		for c2 in range(B):
			tmpC = exp(l[z + c2])
			sumC += tmpC
			l[z + c2] = tmpC
		for c2 in range(B):
			l[z + c2] /= sumC

# Convert log-likes to normalized likes
cpdef void createLikes(
		f32[::1] L, const u32[::1] k_vec, const u32[::1] x_vec
	) noexcept nogil:
	cdef:
		Py_ssize_t W = k_vec.shape[0]
		size_t w
	for w in prange(W, schedule='guided'):
		_normLikes(&L[x_vec[w]], k_vec[w])

# Calculate emission probabilities using hard calls
cpdef void hardEmissions(
		 const u8[:,::1] Z, f64[:,:,::1] E, const f64[::1] P, const u32[::1] c_vec
	) noexcept nogil:
	cdef:
		Py_ssize_t N = E.shape[0]
		Py_ssize_t W = E.shape[1]
		Py_ssize_t K = E.shape[2]
		size_t i, k, w
	for i in prange(N, schedule='guided'):
		for w in range(W):
			p = &P[c_vec[w] + Z[i,w]*K]
			for k in range(K):
				E[i,w,k] = log(p[k])

# Calculate emission probabilities using cluster probabilities
cpdef void softEmissions(
		const u8[:,::1] Z, f64[:,:,::1] E, f64[::1] P, f32[::1] L, const u32[::1] k_vec, const u32[::1] c_vec, 
		const u32[::1] x_vec
	) noexcept nogil:
	cdef:
		Py_ssize_t N = E.shape[0]
		Py_ssize_t W = E.shape[1]
		Py_ssize_t K = E.shape[2]
		Py_ssize_t B
		size_t c, i, k, s, w, x, z
		f32* l
		f64* p
	for i in prange(N, schedule='guided'):
		for w in range(W):
			s = c_vec[w]
			x = x_vec[w]
			B = k_vec[w]
			z = Z[i,w]
			l = &L[x + z*B]
			p = &P[c_vec[w] + z*K]
			for k in range(K):
				for c in range(B):
					E[i,w,k] += <f64>l[c]*p[k]
				E[i,w,k] = log(E[i,w,k])

# Calculate transition probabilities - T[k2, k1] = P(Z_{w} = k1 | Z_{w - 1} = k2)
cpdef void calcTransition(
		f64[:,::1] T, const f64[::1] Q, const f64 a
	) noexcept nogil:
	cdef:
		Py_ssize_t K = T.shape[0]
		size_t k1, k2
		f64 e = exp(-a)
	for k2 in range(K):
		for k1 in range(K):
			if k2 == k1:
				T[k2,k1] = log((1.0 - e)*Q[k1] + e)
			else:
				T[k2,k1] = log((1.0 - e)*Q[k1])

# Viterbi algorithm
cpdef void viterbi(
		const f64[:,::1] E, const f64[::1] Q_log, const f64[:,::1] T, f64[:,::1] A, u8[:,::1] I, u8[::1] V
	) noexcept nogil:
	cdef:
		Py_ssize_t W = E.shape[0]
		Py_ssize_t K = E.shape[1]
		size_t k, w, k1, k2
		f64 max_v
	# Loop through sequence
	for k in range(K):
		A[0,k] = E[0,k] + Q_log[k]
	for w in range(1, W):
		for k1 in range(K):
			A[w,k1] = A[w - 1,0] + T[0,k1]
			I[w,k1] = 0
			for k2 in range(1, K):
				max_v = A[w - 1,k2] + T[k2,k1]
				if max_v > A[w,k1]:
					A[w,k1] = max_v
					I[w,k1] = k2
			A[w,k1] += E[w,k1]
	
	# Decode path
	V[W - 1] = 0
	max_v = A[W - 1,0]
	for k in range(1, K):
		if A[W - 1,k] > max_v:
			V[W - 1] = k
			max_v = A[W - 1,k]
	for w in range(W - 2, -1, -1):
		V[w] = I[w + 1,V[w + 1]]

# Forward-backward algorithm
cpdef void calcFwdBwd(
		const f64[:,::1] E, f64[:,::1] L, const f64[::1] Q_log, const f64[:,::1] T, f64[:,::1] A, f64[:,::1] B, 
		f64[::1] v
	) noexcept nogil:
	cdef:
		Py_ssize_t W = E.shape[0]
		Py_ssize_t K = E.shape[1]
		size_t k, w, k1, k2
		f64 l_fwd
	# Forward calculations
	for k in range(K):
		A[0,k] = E[0,k] + Q_log[k]
	for w in range(1, W):
		for k1 in range(K):
			for k2 in range(K):
				v[k2] = A[w - 1,k2] + T[k2,k1]
			A[w,k1] = _logsumexp(&v[0], K) + E[w,k1]
	
	# Forward log-likelihood
	for k in range(K):
		v[k] = A[W - 1,k]
	l_fwd = _logsumexp(&v[0], K)

	# Backward calculations
	for k in range(K):
		B[W - 1,k] = 0.0
	for w in range(W - 2, -1, -1):
		for k1 in range(K):
			for k2 in range(K):
				v[k2] = B[w + 1,k2] + E[w + 1,k2] + T[k1,k2]
			B[w,k1] = _logsumexp(&v[0], K)

	# Compute posterior probabilities
	for w in range(W):
		for k in range(K):
			L[w,k] = exp(A[w,k] + B[w,k] - l_fwd)



"""
Main caller of hapla.
"""

__author__ = "Jonas Meisner"

# Libraries
import argparse
import sys

# Main function
def main():
	# Argparser
	parser = argparse.ArgumentParser(prog="hapla")
	subparsers = parser.add_subparsers(title="hapla commands")

	### Commands
	# hapla cluster
	parser_cluster = subparsers.add_parser("cluster")
	parser_cluster.add_argument("-g", "--vcf", "--bcf",
		metavar="FILE", help="Input phased genotype file in VCF/BCF format")
	parser_cluster.add_argument("-f", "--size", type=int,
		metavar="INT", help="Use fixed window size")
	parser_cluster.add_argument("-w", "--windows", metavar="FILE",
		help="Use provided window indices")
	parser_cluster.add_argument("-l", "--lmbda", type=float, default=0.125,
		metavar="FLOAT", help="Set lambda hyperparameter (0.125)")
	parser_cluster.add_argument("-s", "--step", type=int,
		metavar="INT", help="Step-size for sliding window")
	parser_cluster.add_argument("-t", "--threads", type=int, default=1,
		metavar="INT", help="Number of threads (1)")
	parser_cluster.add_argument("-o", "--out", default="hapla.cluster",
		metavar="OUTPUT", help="Output prefix")
	parser_cluster.add_argument("--min-freq", type=float, default=0.005,
		metavar="FLOAT", help="Minimum haplotype cluster frequency (0.005)")
	parser_cluster.add_argument("--min-mac", type=int,
		metavar="INT", help="Minimum haplotype cluster allele count")
	parser_cluster.add_argument("--max-clusters", type=int, default=256,
		metavar="INT", help="Maximum number of haplotype clusters per window (256)")
	parser_cluster.add_argument("--max-iterations", type=int, default=1000,
		metavar="INT", help="Maximum number of iterations (1000)")
	parser_cluster.add_argument("--medians", action="store_true",
		help="Save haplotype cluster medians")
	parser_cluster.add_argument("--plink", action="store_true",
		help="Generate binary PLINK output")
	parser_cluster.add_argument("--duplicate-fid", action="store_true",
		help="Use sample list as family ID (PLINK 1.9 compatibility)")
	parser_cluster.add_argument("--memory", action="store_true",
		help="Store haplotypes in 1-bit matrix")

	# hapla struct
	parser_struct = subparsers.add_parser("struct")
	parser_struct.add_argument("-f", "--filelist",
		metavar="FILE", help="Filelist with paths to haplotype cluster alleles files")
	parser_struct.add_argument("-z", "--clusters", metavar="FILE",
		help="Path to a single haplotype cluster alleles file")
	parser_struct.add_argument("-t", "--threads", type=int, default=1,
		metavar="INT", help="Number of threads (1)")
	parser_struct.add_argument("-o", "--out", default="hapla.pca",
		metavar="OUTPUT", help="Output prefix")
	parser_struct.add_argument("--grm", action="store_true",
		help="Estimate genome-wide relationship matrix (GRM)")
	parser_struct.add_argument("--chunk", type=int, default=4096,
		metavar="INT", help="Number of cluster alleles in batches (4096)")
	parser_struct.add_argument("--power", type=int, default=11,
		metavar="INT", help="Number of power iterations to perform (11)")
	parser_struct.add_argument("--no-centering", action="store_true",
		help="Do not perform Gower and data centering of GRM")
	parser_struct.add_argument("--duplicate-fid", action="store_true",
		help="Use sample list as family ID for GCTA format")
	parser_struct.add_argument("--pca", type=int,
		metavar="INT", help="Perform PCA and extract eigenvectors")
	parser_struct.add_argument("--memory", action="store_true",
		help="Perform memory efficient SVD for large-scale data")
	parser_struct.add_argument("--seed", type=int, default=42,
		metavar="INT", help="Random seed (42)")
	parser_struct.add_argument("--loadings", action="store_true",
		help="Save loadings of SVD")
	parser_struct.add_argument("--raw", action="store_true",
		help="Raw output without '*.fam' info")

	# hapla predict
	parser_predict = subparsers.add_parser("predict")
	parser_predict.add_argument("-g", "--vcf", "--bcf",
		metavar="FILE", help="Input phased genotype file in VCF/BCF format")
	parser_predict.add_argument("-r", "--ref", metavar="FILE",
		help="Input reference prefix of pre-estimated cluster medians")
	parser_predict.add_argument("-t", "--threads", type=int, default=1,
		metavar="INT", help="Number of threads (1)")
	parser_predict.add_argument("-o", "--out", default="hapla.predict",
		metavar="OUTPUT", help="Output prefix")
	parser_predict.add_argument("--plink", action="store_true",
		help="Generate binary PLINK output")
	parser_predict.add_argument("--duplicate-fid", action="store_true",
		help="Use sample list as family ID (PLINK 1.9 compatibility)")
	parser_predict.add_argument("--memory", action="store_true",
		help="Store haplotypes in 2-bit matrix")

	# hapla admix
	parser_admix = subparsers.add_parser("admix")
	parser_admix.add_argument("-f", "--filelist",
		metavar="FILE", help="Filelist with paths to haplotype cluster alleles files")
	parser_admix.add_argument("-z", "--clusters",
		metavar="FILE", help="Path to a single haplotype cluster alleles file")
	parser_admix.add_argument("-k", "--K", type=int,
		metavar="INT", help="Number of ancestral components")
	parser_admix.add_argument("-t", "--threads", type=int, default=1,
		metavar="INT", help="Number of threads (1)")
	parser_admix.add_argument("-o", "--out", default="hapla.admix",
		metavar="OUTPUT", help="Output prefix")
	parser_admix.add_argument("--seed", type=int, default=42,
		metavar="INT", help="Random seed (42)")
	parser_admix.add_argument("--iter", type=int, default=1000,
		metavar="INT", help="Maximum number of iterations (1000)")
	parser_admix.add_argument("--tole", type=float, default=1e-9,
		metavar="FLOAT", help="Tolerance in scaled log-likelihood units (1e-9)")
	parser_admix.add_argument("--batches", type=int, default=16,
		metavar="INT", help="Number of initial mini-batches (16)")
	parser_admix.add_argument("--check", type=int, default=5,
		metavar="INT", help="Number of iterations between convergence checks (5)")
	parser_admix.add_argument("--supervised",
		metavar="FILE", help="Path to population assignment file")
	parser_admix.add_argument("--projection",
		metavar="FILE", help="Path to ancestral haplotype cluster allele frequencies file")
	parser_admix.add_argument("--chunk", type=int, default=4096,
		metavar="INT", help="Number of expected cluster alleles in batches (4096)")
	parser_admix.add_argument("--power", type=int, default=11,
		metavar="INT", help="Number of power iterations to perform (11)")
	parser_admix.add_argument("--als-iter", metavar="INT", type=int, default=1000,
		help="Maximum number of iterations in ALS (1000)")
	parser_admix.add_argument("--als-tole", metavar="FLOAT", type=float, default=1e-4,
		help="Tolerance for RMSE of P between iterations (1e-4)")
	parser_admix.add_argument("--subsampling", metavar="INT", type=int, default=4,
		help="Subsampling factor for ALS/SVD initialization (4)")
	parser_admix.add_argument("--no-freqs", action="store_true",
		help="Do not save haplotype cluster frequencies")
	parser_admix.add_argument("--random-init", action="store_true",
		help="Random initialization of parameters")
	parser_admix.add_argument("--prefix", default="chr",
		metavar="OUTPUT", help="Prefix for multiple haplotype cluster frequency files")

	# hapla fatash
	parser_fatash = subparsers.add_parser("fatash")
	parser_fatash.add_argument("-f", "--filelist",
		metavar="FILE", help="Filelist with paths to haplotype cluster alleles files")
	parser_fatash.add_argument("-z", "--clusters",
		metavar="FILE", help="Path to a single haplotype cluster alleles file")
	parser_fatash.add_argument("-e", "--pfilelist",
		metavar="FILE", help="Filelist with paths to haplotype cluster frequencies files")
	parser_fatash.add_argument("-p", "--pfile", 
		metavar="FILE", help="Path to file with haplotype cluster frequencies")
	parser_fatash.add_argument("-q", "--qfile", 
		metavar="FILE", help="Path to file with admixture proportions")
	parser_fatash.add_argument("-t", "--threads", type=int, default=1,
		metavar="INT", help="Number of threads (1)")
	parser_fatash.add_argument("-o", "--out", default="hapla.fatash",
		metavar="OUTPUT", help="Output prefix")
	parser_fatash.add_argument("--alpha", type=float, default=1e-7,
		metavar="FLOAT", help="Set fixed alpha rate (1e-7)")
	parser_fatash.add_argument("--medians", action="store_true",
		help="Utilize haplotype cluster probabilities")
	parser_fatash.add_argument("--viterbi", action="store_true",
		help="Perform Viterbi decoding")
	parser_fatash.add_argument("--save-posterior", action="store_true",
		help="Save posterior probabilities")
	parser_fatash.add_argument("--prefix", default="chr",
		metavar="OUTPUT", help="Prefix for multiple path files")

	# Parse arguments
	args = parser.parse_args()
	if len(sys.argv) < 2:
		parser.print_help()
		sys.exit()

	### Run specified command
	# hapla cluster
	if sys.argv[1] == "cluster":
		if len(sys.argv) < 3:
			parser_cluster.print_help()
			sys.exit()
		else:
			from hapla import cluster
			deaf = vars(parser_cluster.parse_args([]))
			cluster.main(args, deaf)
	
	# hapla struct
	if sys.argv[1] == "struct":
		if len(sys.argv) < 3:
			parser_struct.print_help()
			sys.exit()
		else:
			from hapla import struct
			deaf = vars(parser_struct.parse_args([]))
			struct.main(args, deaf)

	# hapla predict
	if sys.argv[1] == "predict":
		if len(sys.argv) < 3:
			parser_predict.print_help()
			sys.exit()
		else:
			from hapla import predict
			deaf = vars(parser_predict.parse_args([]))
			predict.main(args, deaf)

	# hapla admix
	if sys.argv[1] == "admix":
		if len(sys.argv) < 3:
			parser_admix.print_help()
			sys.exit()
		else:
			from hapla import admix
			deaf = vars(parser_admix.parse_args([]))
			admix.main(args, deaf)

	# hapla fatash
	if sys.argv[1] == "fatash":
		if len(sys.argv) < 3:
			parser_fatash.print_help()
			sys.exit()
		else:
			from hapla import fatash
			deaf = vars(parser_fatash.parse_args([]))
			fatash.main(args, deaf)



##### Define main #####
if __name__ == "__main__":
	main()



# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
cimport numpy as np
from cython.parallel import prange
from libc.stdint cimport uint8_t, uint32_t

ctypedef uint8_t u8
ctypedef uint32_t u32
ctypedef float f32


##### hapla - analyses on haplotype cluster assignments #####
### Inline functions
# Center function
cdef inline void _center(
		const u8* z, f32* x, const f32 u, const Py_ssize_t N
	) noexcept nogil:
	cdef:
		size_t i
	for i in range(N):
		x[i] = <f32>z[i] - u

# Standardize function
cdef inline void _standardize(
		const u8* z, f32* x, const f32 u, const f32 a, const Py_ssize_t N
	) noexcept nogil:
	cdef:
		size_t i
	for i in range(N):
		x[i] = (<f32>z[i] - u)*a

# Calculate Hamming distance
cdef inline u32 hammingPred(
		const u8* X, const u8* R, const Py_ssize_t M
	) noexcept nogil:
	cdef:
		size_t j
		u32 dist = 0
	for j in range(M):
		if X[j] != R[j] and X[j] != 9: # Ignore missing
			dist += 1
	return dist



### hapla struct
# Extract aggregated haplotype cluster counts
cpdef void haplotypeAggregate(
		u8[:,::1] Z, u8[:,::1] Z_agg, f32[::1] p, const u32[::1] k_vec, const u32[::1] c_vec
	) noexcept nogil:
	cdef:
		Py_ssize_t W = Z.shape[0]
		Py_ssize_t N = Z.shape[1]
		size_t c, i, l, s, w
		u8* z
		f32 d = 1.0/<f32>N
	for w in prange(W, schedule='guided'):
		s = c_vec[w]
		for c in range(k_vec[w]):
			l = s + c
			z = &Z_agg[l,0]
			for i in range(N):
				z[i >> 1] += 1 if Z[w,i] == c else 0
				p[l] += 1.0 if Z[w,i] == c else 0.0
			p[l] *= d

# Estimate haplotype cluster frequencies
cpdef void estimateFreq(
		u8[:,::1] Z, f32[::1] p, const u32[::1] k_vec, const u32[::1] c_vec
	) noexcept nogil:
	cdef:
		Py_ssize_t W = Z.shape[0]
		Py_ssize_t N = Z.shape[1]
		size_t c, i, l, s, w
		u8* z
		f32 d = 1.0/<f32>N
	for w in prange(W, schedule='guided'):
		s = c_vec[w]
		z = &Z[w,0]
		for c in range(k_vec[w]):
			l = s + c
			for i in range(N):
				p[l] += 1.0 if z[i] == c else 0.0
			p[l] *= d

# Center expanded batch haplotype cluster assignment matrix
cpdef void centerZ(
		u8[:,::1] Z_agg, f32[:,::1] X, const f32[::1] p, const int S
	) noexcept nogil:
	cdef:
		Py_ssize_t M = X.shape[0]
		Py_ssize_t N = X.shape[1]
		size_t j, l
		f32 u
	for j in prange(M, schedule='guided'):
		l = S + j
		u = 2.0*p[l]
		_center(&Z_agg[l,0], &X[j,0], u, N)

# Standardize expanded chunk of haplotype cluster assignment matrix
cpdef void chunkZ(
		u8[:,::1] Z_agg, f32[:,::1] X, const f32[::1] p, const f32[::1] a
	) noexcept nogil:
	cdef:
		Py_ssize_t M = X.shape[0]
		Py_ssize_t N = X.shape[1]
		size_t j
		f32 d, u
	for j in prange(M, schedule='guided'):
		u = 2.0*p[j]
		_standardize(&Z_agg[j,0], &X[j,0], u, a[j], N)

# Standardize 
cpdef void memoryC(
		u8[:,::1] Z, f32[:,::1] X, const f32[::1] p, const f32[::1] a, const u32[::1] k_vec, const u32[::1] c_vec
	) noexcept nogil:
	cdef:
		Py_ssize_t W = Z.shape[0]
		Py_ssize_t N = X.shape[1]
		size_t c, i, l, s, w
		u8* z
		f32 u, d
		f32* x
	for w in prange(W, schedule='guided'):
		s = c_vec[w] - c_vec[0]
		z = &Z[w,0]
		for c in range(k_vec[w]):
			l = s + c
			u = p[l]
			d = a[l]
			x = &X[l,0]
			for i in range(N):
				x[i] = (1.0 - u)*d if z[2*i] == c else (0.0 - u)*d
				x[i] += (1.0 - u)*d if z[2*i + 1] == c else (0.0 - u)*d



### hapla admix
# Center batch haplotype cluster assignment matrix
cpdef void centerC(
		u8[:,::1] Z, f32[:,::1] X, const f32[::1] p, const u32[::1] k_vec, const u32[::1] c_vec
	) noexcept nogil:
	cdef:
		Py_ssize_t W = Z.shape[0]
		Py_ssize_t N = X.shape[1]
		size_t c, i, l, s, w
		u8* z
		f32 u
		f32* x
	for w in prange(W, schedule='guided'):
		s = c_vec[w] - c_vec[0]
		z = &Z[w,0]
		for c in range(k_vec[w]):
			l = s + c
			u = p[l]
			x = &X[l,0]
			for i in range(N):
				x[i] = (1.0 - u) if z[2*i] == c else (0.0 - u)
				x[i] += (1.0 - u) if z[2*i + 1] == c else (0.0 - u)



### hapla predict
# Haplotype cluster assignment based on pre-estimated medians
cpdef void predictCluster(
		u8[:,::1] X, const u8[:,::1] R, u8[:,::1] Z, const u32 K, const int w
	) noexcept nogil:
	cdef:
		Py_ssize_t N = X.shape[0]
		Py_ssize_t M = X.shape[1]
		size_t c, d, i, k, z
		u8* h
	for i in prange(N):
		h = &X[i,0]
		z = 0
		c = hammingPred(h, &R[0,0], M)
		for k in range(1, K):
			d = hammingPred(h, &R[k,0], M)
			if d <= c:
				z = k
				c = d
		Z[w,i] = z
